import { defineStore } from "pinia";
import {
  BucketService,
  type OwnershipTransferRequestIn,
  type OwnershipTransferRequestOut,
  OwnershipTypeEnum,
  ResourceService,
  WorkflowService,
} from "@/client";
import { useUserStore } from "@/stores/users";
import { useBucketStore } from "@/stores/buckets";
import { useWorkflowStore } from "@/stores/workflows";
import { useResourceStore } from "@/stores/resources";

export const useOTRStore = defineStore({
  id: "otrs",
  state: () =>
    ({
      otrMapping: {},
      __otrLoaded: {
        bucket: false,
        resource: false,
        workflow: false,
      },
    }) as {
      otrMapping: Record<string, OwnershipTransferRequestOut>;
      __otrLoaded: Record<OwnershipTypeEnum, boolean>;
    },
  getters: {
    bucketOtrs(): OwnershipTransferRequestOut[] {
      return Object.values(this.otrMapping).filter(
        (otr) => otr.target_type === OwnershipTypeEnum.BUCKET,
      );
    },
    workflowOtrs(): OwnershipTransferRequestOut[] {
      return Object.values(this.otrMapping).filter(
        (otr) => otr.target_type === OwnershipTypeEnum.WORKFLOW,
      );
    },
    resourceOtrs(): OwnershipTransferRequestOut[] {
      return Object.values(this.otrMapping).filter(
        (otr) => otr.target_type === OwnershipTypeEnum.RESOURCE,
      );
    },
  },
  actions: {
    __filterOTRMapping(type: OwnershipTypeEnum) {
      for (const otr_id of Object.keys(this.otrMapping)) {
        if (this.otrMapping[otr_id].target_type === type) {
          delete this.otrMapping[otr_id];
        }
      }
    },
    fetchOtrs(
      targetType: OwnershipTypeEnum,
      currentOwnerId?: string,
      newOwnerId?: string,
    ): Promise<OwnershipTransferRequestOut[]> {
      switch (targetType) {
        case OwnershipTypeEnum.BUCKET:
          return BucketService.bucketListBucketOtrs(currentOwnerId, newOwnerId);
        case OwnershipTypeEnum.WORKFLOW:
          return WorkflowService.workflowListWorkflowOtrs(
            currentOwnerId,
            newOwnerId,
          );
        case OwnershipTypeEnum.RESOURCE:
          return ResourceService.resourceListResourceOtrs(
            currentOwnerId,
            newOwnerId,
          );
      }
    },
    fetchOwnOtrs(
      targetType: OwnershipTypeEnum,
      onFinally?: () => void,
    ): Promise<OwnershipTransferRequestOut[]> {
      if (this.__otrLoaded[targetType]) {
        onFinally?.();
      }
      const userRepo = useUserStore();
      return Promise.all([
        this.fetchOtrs(targetType, userRepo.currentUID),
        this.fetchOtrs(targetType, undefined, userRepo.currentUID),
      ])
        .then((otrs) => otrs.flat())
        .then((otrs) => {
          this.__filterOTRMapping(targetType);
          for (const otr of otrs) {
            this.otrMapping[otr.target_id] = otr;
          }
          userRepo.fetchUsernames(
            otrs
              .map((otr) => [otr.new_owner_uid, otr.current_owner_uid ?? ""])
              .flat()
              .filter((uid) => uid.length > 0),
          );
          return otrs;
        })
        .finally(() => {
          this.__otrLoaded[targetType] = true;
          onFinally?.();
        });
    },
    acceptOtr(otr: OwnershipTransferRequestOut): Promise<string> {
      const promise = (function () {
        switch (otr.target_type) {
          case OwnershipTypeEnum.BUCKET:
            return useBucketStore()
              .acceptBucketOtr(otr)
              .then((bucket) => {
                return bucket.name;
              });
          case OwnershipTypeEnum.WORKFLOW:
            return useWorkflowStore()
              .acceptWorkflowOtr(otr)
              .then((workflow) => {
                return workflow.workflow_id;
              });
          case OwnershipTypeEnum.RESOURCE:
            return useResourceStore()
              .acceptResourceOtr(otr)
              .then((resource) => {
                return resource.resource_id;
              });
        }
      })();
      return promise.then((target_id) => {
        delete this.otrMapping[otr.target_id];
        return target_id;
      });
    },
    deleteOtr(otr: OwnershipTransferRequestOut): Promise<void> {
      const promise = (function () {
        switch (otr.target_type) {
          case OwnershipTypeEnum.BUCKET:
            return BucketService.bucketDeleteBucketOtr(otr.target_id);
          case OwnershipTypeEnum.WORKFLOW:
            return WorkflowService.workflowDeleteWorkflowOtr(otr.target_id);
          case OwnershipTypeEnum.RESOURCE:
            return ResourceService.resourceDeleteResourceOtr(otr.target_id);
        }
      })();
      return promise.then((target_id) => {
        delete this.otrMapping[otr.target_id];
        return target_id;
      });
    },
    createOtr(
      targetId: string,
      otr: OwnershipTransferRequestIn,
      targetType: OwnershipTypeEnum,
    ): Promise<OwnershipTransferRequestOut> {
      const promise = (function () {
        switch (targetType) {
          case OwnershipTypeEnum.BUCKET:
            return BucketService.bucketCreateBucketOtr(targetId, otr);
          case OwnershipTypeEnum.WORKFLOW:
            return WorkflowService.workflowCreateWorkflowOtr(targetId, otr);
          case OwnershipTypeEnum.RESOURCE:
            return ResourceService.resourceCreateResourceOtr(targetId, otr);
        }
      })();
      return promise.then((newOtr) => {
        this.otrMapping[newOtr.target_id] = newOtr;
        return newOtr;
      });
    },
  },
});
